import time
import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader

from sgl.tasks.base_task import BaseTask
from sgl.tasks.utils import accuracy, set_seed, train, mini_batch_train, evaluate, mini_batch_evaluate


class NodeClassification(BaseTask):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, loss_fn=nn.CrossEntropyLoss(), seed=42,
                 train_batch_size=None, eval_batch_size=None):
        super(NodeClassification, self).__init__()

        self.__dataset = dataset
        self.__labels = self.__dataset.y

        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr,
                                weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed

        self.__mini_batch = False
        if train_batch_size is not None:
            self.__mini_batch = True
            self.__train_loader = DataLoader(
                self.__dataset.train_idx, batch_size=train_batch_size, shuffle=True, drop_last=False)
            self.__val_loader = DataLoader(
                self.__dataset.val_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__test_loader = DataLoader(
                self.__dataset.test_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__all_eval_loader = DataLoader(
                range(self.__dataset.data.num_node), batch_size=eval_batch_size, shuffle=False, drop_last=False)

        self.__test_acc = self._execute()

    @property
    def test_acc(self):
        return self.__test_acc

    def _execute(self):
        set_seed(self.__seed)

        pre_time_st = time.time()
        self.__model.preprocess(self.__dataset.adj, self.__dataset.x)
        pre_time_ed = time.time()
        print(f"Preprocessing done in {(pre_time_ed - pre_time_st):.4f}s")

        self.__model = self.__model.to(self.__device)
        self.__labels = self.__labels.to(self.__device)

        t_total = time.time()
        best_val = 0.
        best_test = 0.
        for epoch in range(self.__epochs):
            t = time.time()
            if self.__mini_batch is False:
                loss_train, acc_train = train(self.__model, self.__dataset.train_idx, self.__labels, self.__device,
                                              self.__optimizer, self.__loss_fn)
                acc_val, acc_test = evaluate(self.__model, self.__dataset.val_idx, self.__dataset.test_idx,
                                             self.__labels, self.__device)
            else:
                loss_train, acc_train = mini_batch_train(self.__model, self.__dataset.train_idx, self.__train_loader,
                                                         self.__labels, self.__device, self.__optimizer, self.__loss_fn)
                acc_val, acc_test = mini_batch_evaluate(self.__model, self.__dataset.val_idx, self.__val_loader,
                                                        self.__dataset.test_idx, self.__test_loader, self.__labels,
                                                        self.__device)

            print('Epoch: {:03d}'.format(epoch + 1),
                  'loss_train: {:.4f}'.format(loss_train),
                  'acc_train: {:.4f}'.format(acc_train),
                  'acc_val: {:.4f}'.format(acc_val),
                  'acc_test: {:.4f}'.format(acc_test),
                  'time: {:.4f}s'.format(time.time() - t))
            if acc_val > best_val:
                best_val = acc_val
                best_test = acc_test

        acc_val, acc_test = self._postprocess()
        if acc_val > best_val:
            best_val = acc_val
            best_test = acc_test

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')
        return best_test

    def _postprocess(self):
        self.__model.eval()
        if self.__mini_batch is False:
            outputs = self.__model.model_forward(
                range(self.__dataset.num_node), self.__device).to("cpu")
        else:
            outputs = None
            for batch in self.__all_eval_loader:
                output = self.__model.model_forward(batch, self.__device)
                if outputs is None:
                    outputs = output
                else:
                    outputs = torch.vstack((outputs, output))

        final_output = self.__model.postprocess(self.__dataset.adj, outputs)
        acc_val = accuracy(
            final_output[self.__dataset.val_idx], self.__labels[self.__dataset.val_idx])
        acc_test = accuracy(
            final_output[self.__dataset.test_idx], self.__labels[self.__dataset.test_idx])
        return acc_val, acc_test


class HeteroNodeClassification(BaseTask):
    def __init__(self, dataset, predict_class, model, lr, weight_decay, epochs,
                 device, loss_fn=nn.CrossEntropyLoss(), seed=42,
                 train_batch_size=None, eval_batch_size=None,
                 random_subgraph_num=-1, subgraph_edge_type_num=-1,
                 subgraph_list=None, record_subgraph_weight=False):

        if subgraph_list is None and (random_subgraph_num == -1 or subgraph_edge_type_num == -1):
            raise ValueError(
                "Either subgraph_list or (random_subgraph_num, subgraph_edge_type_num) should be provided!")
        if subgraph_list is not None and (random_subgraph_num != -1 or subgraph_edge_type_num != -1):
            raise ValueError(
                "subgraph_list is provided, random_subgraph_num and subgraph_edge_type_num will be ignored!")

        super(HeteroNodeClassification, self).__init__()

        self.__dataset = dataset
        self.__predict_class = predict_class
        self.__labels = self.__dataset.data[self.__predict_class].y.squeeze(-1)

        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr,
                                weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed

        self.__mini_batch = False
        if train_batch_size is not None:
            self.__mini_batch = True
            self.__train_loader = DataLoader(
                self.__dataset.train_idx, batch_size=train_batch_size, shuffle=True, drop_last=False)
            self.__val_loader = DataLoader(
                self.__dataset.val_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__test_loader = DataLoader(
                self.__dataset.test_idx, batch_size=eval_batch_size, shuffle=False, drop_last=False)

        self.__test_acc, self.__subgraph_weight = self._execute(
            random_subgraph_num, subgraph_edge_type_num,
            subgraph_list, record_subgraph_weight)

    @property
    def test_acc(self):
        return self.__test_acc

    @property
    def subgraph_weight(self):
        return self.__subgraph_weight

    def _execute(self, random_subgraph_num=-1, subgraph_edge_type_num=-1,
                 subgraph_list=None, record_subgraph_weight=False):
        set_seed(self.__seed)

        pre_time_st = time.time()
        self.__model.preprocess(
            self.__dataset, self.__predict_class,
            random_subgraph_num, subgraph_edge_type_num, subgraph_list)
        pre_time_ed = time.time()
        print(f"Preprocessing done in {(pre_time_ed - pre_time_st):.4f}s")

        self.__model = self.__model.to(self.__device)
        self.__labels = self.__labels.to(self.__device)

        t_total = time.time()
        best_val = 0.
        best_test = 0.
        if record_subgraph_weight:
            best_subgraph_weight = self.__model.subgraph_weight.clone()
        for epoch in range(self.__epochs):
            t = time.time()
            if self.__mini_batch is False:
                loss_train, acc_train = train(self.__model, self.__dataset.train_idx, self.__labels, self.__device,
                                              self.__optimizer, self.__loss_fn)
                acc_val, acc_test = evaluate(self.__model, self.__dataset.val_idx, self.__dataset.test_idx,
                                             self.__labels, self.__device)
            else:
                loss_train, acc_train = mini_batch_train(self.__model, self.__dataset.train_idx, self.__train_loader,
                                                         self.__labels, self.__device, self.__optimizer, self.__loss_fn)
                acc_val, acc_test = mini_batch_evaluate(self.__model, self.__dataset.val_idx, self.__val_loader,
                                                        self.__dataset.test_idx, self.__test_loader, self.__labels,
                                                        self.__device)

            print('Epoch: {:03d}'.format(epoch + 1),
                  'loss_train: {:.4f}'.format(loss_train),
                  'acc_train: {:.4f}'.format(acc_train),
                  'acc_val: {:.4f}'.format(acc_val),
                  'acc_test: {:.4f}'.format(acc_test),
                  'time: {:.4f}s'.format(time.time() - t))
            if acc_val > best_val:
                best_val = acc_val
                best_test = acc_test
                if record_subgraph_weight:
                    best_subgraph_weight = self.__model.subgraph_weight.clone()

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')

        if record_subgraph_weight:
            return best_test, best_subgraph_weight
        else:
            return best_test, None



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import MultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import IterateLearnableWeightedMessageOp


class GAMLPRecursive(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers):
        super(GAMLPRecursive, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = IterateLearnableWeightedMessageOp(0, prop_steps + 1, "recursive", feat_dim)
        self._base_model = MultiLayerPerceptron(feat_dim, hidden_dim, num_layers, output_dim)



import torch
from torch import Tensor

from sgl.operators.base_op import MessageOp
from sgl.operators.utils import one_dim_weighted_add


class SimpleWeightedMessageOp(MessageOp):

    # 'alpha' needs one additional parameter 'alpha';
    # 'hand_crafted' needs one additional parameter 'weight_list'
    def __init__(self, start, end, combination_type, *args):
        super(SimpleWeightedMessageOp, self).__init__(start, end)
        self._aggr_type = "simple_weighted"

        if combination_type not in ["alpha", "hand_crafted"]:
            raise ValueError(
                "Invalid weighted combination type! Type must be 'alpha' or 'hand_crafted'.")
        self.__combination_type = combination_type

        if len(args) != 1:
            raise ValueError(
                "Invalid parameter numbers for the simple weighted aggregator!")
        self.__alpha, self.__weight_list = None, None
        if combination_type == "alpha":
            self.__alpha = args[0]
            if not isinstance(self.__alpha, float):
                raise TypeError("The alpha must be a float!")
            elif self.__alpha > 1 or self.__alpha < 0:
                raise ValueError("The alpha must be a float in [0,1]!")

        elif combination_type == "hand_crafted":
            self.__weight_list = args[0]
            if isinstance(self.__weight_list, list):
                self.__weight_list = torch.FloatTensor(self.__weight_list)
            elif not isinstance(self.__weight_list, (list, Tensor)):
                raise TypeError(
                    "The input weight list must be a list or a tensor!")

    def _combine(self, feat_list):
        if self.__combination_type == "alpha":
            self.__weight_list = [self.__alpha]
            for _ in range(len(feat_list) - 1):
                self.__weight_list.append(
                    (1 - self.__alpha) * self.__weight_list[-1])
            self.__weight_list = torch.FloatTensor(
                self.__weight_list[self._start:self._end])

        elif self.__combination_type == "hand_crafted":
            pass
        else:
            raise NotImplementedError

        weighted_feat = one_dim_weighted_add(
            feat_list[self._start:self._end], weight_list=self.__weight_list)
        return weighted_feat



import numpy as np
import torch
from scipy.sparse import csr_matrix
from torch import Tensor


# Base class for adjacency matrix
class Edge:
    def __init__(self, row, col, edge_weight, edge_type, num_node, edge_attrs=None):
        if not isinstance(edge_type, str):
            raise TypeError("Edge type must be a string!")
        self.__edge_type = edge_type

        if (not isinstance(row, (list, np.ndarray, Tensor))) or (not isinstance(col, (list, np.ndarray, Tensor))) or (
                not isinstance(edge_weight, (list, np.ndarray, Tensor))):
            raise TypeError("Row, col and edge_weight must be a list, np.ndarray or Tensor!")
        self.__row = row
        self.__col = col
        self.__edge_weight = edge_weight
        self.__edge_attrs = edge_attrs
        self.__num_edge = len(row)

        if isinstance(row, Tensor) or isinstance(col, Tensor):
            self.__sparse_matrix = csr_matrix((edge_weight.numpy(), (row.numpy(), col.numpy())),
                                              shape=(num_node, num_node))
        else:
            self.__sparse_matrix = csr_matrix((edge_weight, (row, col)), shape=(num_node, num_node))

    @property
    def sparse_matrix(self):
        return self.__sparse_matrix

    @property
    def edge_type(self):
        return self.__edge_type

    @property
    def num_edge(self):
        return self.__num_edge

    @property
    def edge_index(self):
        return self.__row, self.__col

    @property
    def edge_attrs(self):
        return self.__edge_attrs

    @edge_attrs.setter
    def edge_attrs(self, edge_attrs):
        # more restrictions

        self.__edge_attrs = edge_attrs

    @property
    def row(self):
        return self.__row

    @property
    def col(self):
        return self.__col

    @property
    def edge_weight(self):
        return self.__edge_weight


# Base class or storing node information
class Node:
    def __init__(self, node_type, num_node, x=None, y=None, node_ids=None):
        if not isinstance(num_node, int):
            raise TypeError("Num nodes must be a integer!")
        elif not isinstance(node_type, str):
            raise TypeError("Node type must be a string!")
        elif (node_ids is not None) and (not isinstance(node_ids, (list, np.ndarray, Tensor))):
            raise TypeError("Node IDs must be a list, np.ndarray or Tensor!")
        self.__num_node = num_node
        self.__node_type = node_type
        if node_ids is not None:
            self.__node_ids = node_ids
        else:
            self.__node_ids = range(num_node)
        self.__x = x
        self.__y = y

    @property
    def num_node(self):
        return self.__num_node

    @property
    def node_ids(self):
        return self.__node_ids

    @property
    def node_type(self):
        return self.__node_type

    @property
    def x(self):
        return self.__x

    @x.setter
    def x(self, x):
        # more restrictions

        self.__x = x

    @property
    def y(self):
        return self.__y

    @y.setter
    def y(self, y):
        # more restrictions

        self.__y = y


# Base class for homogeneous graph
class Graph:
    def __init__(self, row, col, edge_weight, num_node, node_type, edge_type, x=None, y=None, node_ids=None,
                 edge_attr=None):
        self.__edge = Edge(row, col, edge_weight, edge_type, num_node, edge_attr)
        if node_ids is None:
            self.__node_ids = range(num_node)
        else:
            self.__node_ids = node_ids
        self.__node = Node(node_type, num_node, x, y, node_ids)

    @property
    def num_node(self):
        return self.__node.num_node

    @property
    def num_edge(self):
        return self.__edge.num_edge

    @property
    def adj(self):
        return self.__edge.sparse_matrix

    @property
    def edge_type(self):
        return self.__edge.edge_type

    @property
    def node_type(self):
        return self.__node.node_type

    @property
    def x(self):
        return self.__node.x

    @x.setter
    def x(self, x):
        self.__node.x = x

    @property
    def y(self):
        return self.__node.y

    @y.setter
    def y(self, y):
        self.__node.y = y

    @property
    def num_features(self):
        if self.__node.x is not None:
            return self.__node.x.shape[1]

    @property
    def num_classes(self):
        if self.__node.y is not None:
            return int(self.__node.y.max() + 1)

    @property
    def node_degrees(self):
        row_sum = self.adj.sum(axis=1)
        return torch.LongTensor(row_sum).squeeze(1)


# Base class for heterogeneous graph
class HeteroGraph:
    def __init__(self, row_dict, col_dict, edge_weight_dict, num_node_dict, node_types, edge_types, node_id_dict,
                 x_dict=None, y_dict=None, edge_attr_dict=None):
        self.__edges_dict = {}
        self.__edge_types = edge_types
        for edge_type in edge_types:
            if not isinstance(edge_type, str):
                raise TypeError("Edge type must be a string!")
        if (not isinstance(row_dict, dict)) or (not isinstance(col_dict, dict)) or (
                not isinstance(edge_weight_dict, dict)) or (
                edge_attr_dict is not None and not isinstance(edge_attr_dict, dict)):
            raise TypeError("Rows, cols, edge weights and edge attrs must be dicts!")
        elif not isinstance(edge_types, list):
            raise TypeError("Edge types must be a list!")
        elif not ((row_dict.keys() == col_dict.keys()) and (col_dict.keys() == edge_weight_dict.keys()) and (
                list(edge_weight_dict.keys()).sort() == edge_types.copy().sort())):
            raise ValueError("The keys of the rows, cols, edge_weights and edge_types must be the same!")

        for edge_type in edge_types:
            if edge_attr_dict is not None:
                self.__edges_dict[edge_type] = Edge(row_dict[edge_type], col_dict[edge_type],
                                                    edge_weight_dict[edge_type], edge_type,
                                                    edge_attr_dict.get(edge_type, None))
            else:
                self.__edges_dict[edge_type] = Edge(row_dict[edge_type], col_dict[edge_type],
                                                    edge_weight_dict[edge_type], edge_type)

        self.__nodes_dict = {}
        self.__node_types = node_types
        for node_type in node_types:
            if not isinstance(node_type, str):
                raise TypeError("Node type must be a string!")
        if not isinstance(num_node_dict, dict):
            raise TypeError("Num nodes must be a dict!")
        elif not isinstance(node_types, list):
            raise TypeError("Node types must be a list!")
        elif list(num_node_dict.keys()).sort() != node_types.copy().sort():
            raise TypeError("The keys of num_nodes and node_types must be the same!")
        elif ((x_dict is not None) and (not isinstance(x_dict, dict))) or (
                (y_dict is not None) and (not isinstance(y_dict, dict))):
            raise TypeError("Xs and Ys must be a dict!")

        self.__node_id_offsets = {}
        node_count = 0
        for node_type in node_types:
            self.__node_id_offsets[node_type] = node_count
            node_count += num_node_dict[node_type]

        if node_id_dict is None:
            self.__node_id_dict = {}
            for node_type in node_types:
                self.__node_id_dict[node_type] = list(range(self.__node_id_offsets[node_type],
                                                            self.__node_id_offsets[node_type] + num_node_dict[
                                                                node_type]))
        else:
            self.__node_id_dict = node_id_dict

        for node_type in node_types:
            self.__nodes_dict[node_type] = Node(node_type, num_node_dict[node_type], x_dict.get(node_type, None),
                                                y_dict.get(node_type, None), self.__node_id_dict[node_type])

    def __getitem__(self, key):
        if key in self.__edge_types:
            return self.__edges_dict[key]
        elif key in self.__node_types:
            return self.__nodes_dict[key]
        else:
            raise ValueError("Please input valid edge type or node type!")

    def __setitem__(self, key, value):
        if not isinstance(key, str):
            raise TypeError("Edge type or node type must be a string!")
        if key in self.__edge_types:
            if not isinstance(value, Edge):
                raise TypeError("Please organize the dataset using the Edge class!")
            # more restrictions

            self.__edges_dict[key] = value
        elif key in self.__node_types:
            if not isinstance(value, Node):
                raise TypeError("Please organize the dataset using the Node class!")
            # more restrictions

            self.__nodes_dict[key] = value
        else:
            raise ValueError("Please input valid edge type or node type!")

    @property
    def node_id_dict(self):
        return self.__node_id_dict

    @property
    def node_types(self):
        return self.__node_types

    @property
    def edge_types(self):
        return self.__edge_types

    @property
    def num_features(self):
        num_features = {}
        for node_type in self.__node_types:
            x_temp = self.__nodes_dict[node_type].x
            if x_temp is not None:
                num_features[node_type] = x_temp.shape[1]
            else:
                num_features[node_type] = 0
        return num_features

    @property
    def num_classes(self):
        num_classes = {}
        for node_type in self.__node_types:
            if self.__nodes_dict[node_type].y is not None:
                num_classes[node_type] = (self.__nodes_dict[node_type].y.max() + 1)
        return num_classes

    @property
    def num_node(self):
        num_node = {}
        for node_type in self.__node_types:
            num_node[node_type] = self.__nodes_dict[node_type].num_node
        return num_node



import time
import torch
from torch.optim import Adam
from torch.utils.data import DataLoader
import torch.utils.data as Data
import torch.nn.functional as F
import scipy.sparse as sp

from sgl.tasks.base_task import BaseTask
from sgl.tasks.utils import set_seed, edge_predict_eval, mask_test_edges, edge_predict_train, \
    mini_batch_edge_predict_train, mini_batch_edge_predict_eval, edge_predict_score, mix_pos_neg_edges
from sgl.tasks.utils import sparse_mx_to_torch_sparse_tensor, adj_to_symmetric_norm

class LinkPredictionGAE(BaseTask):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, loss_fn=F.binary_cross_entropy_with_logits, seed=42,
                 train_batch_size=None, eval_batch_size=None, threshold=0.5):
        super(LinkPredictionGAE, self).__init__()

        self.__dataset = dataset
        self.__edge_labels = self._generate_edge_labels() # not used yet, maybe in the future
        
        self.__train_adj, self.__train_edges, self.__train_edges_neg, self.__val_edges, self.__val_edges_neg, self.__test_edges, self.__test_edges_neg = \
            mask_test_edges(self.__dataset.adj)
        print("Edge split finished!")
        
        self.__all_edges = torch.cat((self.__train_edges, self.__val_edges, self.__test_edges))
        self.__all_edges_neg = torch.cat((self.__train_edges_neg, self.__val_edges_neg, self.__test_edges_neg))

        self.__model = model
        self.__with_params = len(list(model.parameters())) != 0 
        self.__optimizer = Adam(model.parameters(), lr=lr, weight_decay=weight_decay) if self.__with_params is True else None
        
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed
        self.__pred_threshold = threshold

        self.__mini_batch = False
        if train_batch_size is not None:
            self.__mini_batch = True

            train_all_edges, train_all_lables = mix_pos_neg_edges(self.__train_edges, self.__train_edges_neg, train_batch_size // 2)
            train_dataset = Data.TensorDataset(train_all_edges, train_all_lables)

            val_all_edges, val_all_lables = mix_pos_neg_edges(self.__val_edges, self.__val_edges_neg, eval_batch_size // 2)
            val_dataset = Data.TensorDataset(val_all_edges, val_all_lables)

            test_all_edges, test_all_lables = mix_pos_neg_edges(self.__test_edges, self.__test_edges_neg, eval_batch_size // 2)
            test_dataset = Data.TensorDataset(test_all_edges, test_all_lables)

            all_edges, all_edges_label = mix_pos_neg_edges(self.__all_edges, self.__all_edges_neg, eval_batch_size // 2)
            all_dataset = Data.TensorDataset(all_edges, all_edges_label)

            self.__train_loader = DataLoader(
                dataset=train_dataset, batch_size=train_batch_size, shuffle=True, drop_last=False)
            self.__val_loader = DataLoader(
                dataset=val_dataset, batch_size=eval_batch_size, shuffle=False, drop_last=False)
            self.__test_loader = DataLoader(
                dataset=test_dataset, batch_size=eval_batch_size, shuffle=False, drop_last=False)

            # For post_process evaluation, now is not used
            self.__all_eval_loader = DataLoader(
                dataset=all_dataset, batch_size=eval_batch_size, shuffle=False, drop_last=False)

        self.__test_roc_auc, self.__test_avg_prec = self._execute()

    @property
    def test_roc_auc(self):
        return self.__test_roc_auc

    @property
    def test_avg_prec(self):
        return self.__test_avg_prec

    def _generate_edge_labels(self):
        adj = (self.__dataset.adj + sp.eye(self.__dataset.num_node)).toarray()
        adj_label = torch.FloatTensor(adj)
        return adj_label

    def _execute(self):
        set_seed(self.__seed)

        pre_time_st = time.time()
        self.__model.preprocess(self.__train_adj, self.__dataset.x)
        pre_time_ed = time.time()
        print(f"Preprocessing done in {(pre_time_ed - pre_time_st):.4f}s")

        self.__model = self.__model.to(self.__device)
        self.__edge_labels = self.__edge_labels.to(self.__device)

        train_node_index = torch.arange(0, self.__dataset.num_node, dtype=torch.long)

        t_total = time.time()
        best_roc_auc_val, best_avg_prec_val = 0. , 0.
        best_roc_auc_test, best_avg_prec_test = 0. , 0.

        for epoch in range(self.__epochs):
            t = time.time()
            if self.__mini_batch is False:
                loss_train, roc_auc_train, avg_prec_train = edge_predict_train(self.__model, train_node_index, self.__with_params, self.__train_edges, self.__train_edges_neg,
                                                                               self.__device, self.__optimizer, self.__loss_fn, self.__pred_threshold)
                roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test = edge_predict_eval(self.__model, train_node_index, self.__val_edges, self.__val_edges_neg,
                                                                                           self.__test_edges, self.__test_edges_neg, self.__device, self.__pred_threshold)
            else:
                loss_train, roc_auc_train, avg_prec_train = mini_batch_edge_predict_train(self.__model, train_node_index, self.__with_params, self.__train_loader, self.__device, 
                                                                                    self.__optimizer, self.__loss_fn, self.__pred_threshold)
                roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test = mini_batch_edge_predict_eval(self.__model, train_node_index, self.__val_loader, self.__test_loader, self.__device, self.__pred_threshold)
            print('Epoch: {:03d}'.format(epoch + 1),
                  'loss_train: {:.4f}'.format(loss_train),
                  'roc_auc_train: {:.4f}'.format(roc_auc_train),
                  'avg_prec_train: {:.4f}'.format(avg_prec_train),
                  '\n',
                  'roc_auc_val: {:.4f}'.format(roc_auc_val),
                  'avg_prec_val: {:.4f}'.format(avg_prec_val),
                  'roc_auc_test: {:.4f}'.format(roc_auc_test),
                  'avg_prec_test: {:.4f}'.format(avg_prec_test),
                  'time: {:.4f}s'.format(time.time() - t))

            if roc_auc_val > best_roc_auc_val:
                best_roc_auc_val = roc_auc_val
                best_roc_auc_test = roc_auc_test
            
            if avg_prec_val > best_avg_prec_val:
                best_avg_prec_val = avg_prec_val
                best_avg_prec_test = avg_prec_test

        roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test = self._postprocess()
        if roc_auc_val > best_roc_auc_val:
            best_roc_auc_val = roc_auc_val
            best_roc_auc_test = roc_auc_test
        if avg_prec_val > best_avg_prec_val:
            best_avg_prec_val = avg_prec_val
            best_avg_prec_test = avg_prec_test

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best roc_auc_val: {best_roc_auc_val:.4f}, best roc_auc_test: {best_roc_auc_test:.4f}')
        print(f"Best avg_prec_val: {best_avg_prec_val:.4f}, best avg_prec_test: {best_avg_prec_test:.4f}")
        return best_roc_auc_test, best_avg_prec_test

    def _postprocess(self):
        self.__model.eval()
        if self.__mini_batch is False:
            node_features = self.__model.model_forward(
                range(self.__dataset.num_node), self.__device).to('cpu')
        else:
            # postprocess does not support batch training yet
            outputs = None
            return 0., 0., 0., 0.

        final_node_features = self.__model.postprocess(self.__train_adj, node_features)
        edge_feature = torch.mm(final_node_features, final_node_features.t()).data

        roc_auc_val, avg_prec_val = edge_predict_score(edge_feature, self.__val_edges, self.__val_edges_neg, self.__pred_threshold)
        roc_auc_test, avg_prec_test = edge_predict_score(edge_feature, self.__test_edges, self.__all_edges_neg, self.__pred_threshold)
        return roc_auc_val, avg_prec_val, roc_auc_test, avg_prec_test

class LinkPredictionNAFS(BaseTask):
    """ LinkPrediction Tasks based on NAFS
        Fast and Does Not need train
    """
    def __init__(self, dataset, hops=20, method='mean', seed=42, r_list=[0.5, 0.4, 0.3, 0.2, 0.1, 0], threshold=0.5):
        # NOTE that: hops can be either an integer or a list
        # when it's a list, it contains the numbers of hops that you want to test
        # when it's a numebr, it will test all the hops in range(hops) 
        method = method.lower()
        if method not in ['mean', 'max', 'concat', 'simple']:
            raise ValueError("Method not Suppoted! Choose 'mean', 'max' or 'concat' !")  
        if type(hops) not in [list, int]:
            raise ValueError("hops type not supported!")

        super(LinkPredictionNAFS, self).__init__()
        self.__dataset = dataset

        self.__train_adj, _, _, _, _, self.__test_edges, self.__test_edges_neg = mask_test_edges(self.__dataset.adj)
        print("Edge split finished!")

        self.__method = method
        self.__r_list = r_list
        self.__hops = range(hops) if type(hops) == int else hops
        self.__seed = seed
        self.__pred_threshold = threshold

        self.__best_hop_roc_auc, self.__best_hop_avg_prec, self.__test_roc_auc, self.__test_avg_prec = self._execute()
    
    @property
    def test_roc_auc(self):
        return self.__test_roc_auc

    @property
    def test_avg_prec(self):
        return self.__test_avg_prec

    @property
    def best_hop_roc_auc(self):
        return self.__best_hop_roc_auc

    @property
    def best_hop_avg_prec(self):
        return self.__best_hop_avg_prec

    def _execute(self):
        set_seed(self.__seed)

        t_total = time.time()
        best_roc_auc, best_avg_prec = 0., 0. 
        best_hop_roc_auc, best_hop_avg_prev = 0, 0

        for hop in self.__hops:
            t = time.time()
            roc_auc, avg_prec = self._k_hop_link_prediction(hop)

            print('hops:{:2d}'.format(hop),
                  'roc_auc_score: {:.4f}'.format(roc_auc),
                  'avg_precision: {:.4f}'.format(avg_prec),
                  'time: {:.4f} seconds'.format(time.time() - t)
                  )
            
            if roc_auc > best_roc_auc:
                best_roc_auc = roc_auc
                best_hop_roc_auc = hop
            if avg_prec > best_avg_prec:
                best_avg_prec = avg_prec
                best_hop_avg_prev = hop
            
        print("Node Smoothing Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print("best_roc_auc_score: {:.4f}, best_avg_precision: {:.4f}".format(best_roc_auc, best_avg_prec))

        return best_hop_roc_auc, best_hop_avg_prev, best_roc_auc, best_avg_prec
        
    def _k_hop_link_prediction(self, hops):
        input_features = []
        features = torch.tensor(self.__dataset.x, dtype=torch.float)
        r_list = self.__r_list
        for r in r_list:
            t = time.time()
            adj_norm = adj_to_symmetric_norm(self.__train_adj, r)
            adj_norm = sparse_mx_to_torch_sparse_tensor(adj_norm)
            features_list = []
            features_list.append(features)
            for _ in range(hops):
                features_list.append(torch.spmm(adj_norm, features_list[-1]))

            if self.__method == 'simple':
                input_features.append(features_list[-1])
                break

            weight_list = []
            norm_fea = torch.norm(features, 2, 1).add(1e-10)
            for fea in features_list:
                norm_cur = torch.norm(fea, 2, 1).add(1e-10)
                tmp = torch.div((features * fea).sum(1), norm_cur)
                tmp = torch.div(tmp, norm_fea)

                weight_list.append(tmp.unsqueeze(-1))

            tmp_weight = torch.cat(weight_list, dim=1)
            weight = F.softmax(tmp_weight, dim=1)

            input_feas = []
            for i in range(self.__dataset.num_node):
                fea = 0.
                for j in range(hops + 1):
                    fea += (weight[i][j]*features_list[j][i]).unsqueeze(0)
                input_feas.append(fea)
            input_feas = torch.cat(input_feas, dim=0)
            input_features.append(input_feas)
            print("hops {:d}: r={:.2f} calculation finished in {:.4f} seconds".format(hops, r, time.time() - t))
        
        if self.__method == 'mean':
            input_features = sum(input_features) / len(input_features)
        elif self.__method == 'max':
            input_features =  list(map(lambda x: x.unsqueeze(0), input_features))
            input_features = torch.cat(input_features, dim=0).max(0)[0]
        elif self.__method == 'concat':
            input_features = torch.cat(input_features, dim=1)
        elif self.__method == 'simple':
            input_features = input_features[-1]

        sim = torch.mm(input_features, input_features.t())
        roc_auc, avg_prec = edge_predict_score(sim, self.__test_edges, self.__test_edges_neg, self.__pred_threshold)
        return roc_auc, avg_prec



import numpy as np
import os.path as osp
import pickle as pkl
import torch
from torch_sparse import SparseTensor, coalesce

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


class Actor(NodeDataset):
    # Have 10 different split of training and validation set, identified by split_id in [0, 9]
    # Currently, we only support calculating the accuracy of one split, 
    # and average accuracy will be supported in the future.
    def __init__(self, name="actor", root="./", split="official", split_id=0):
        if split_id not in range(10):
            raise ValueError("Split id not supported")
        super(Actor, self).__init__(root + "Actor", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split, self._split_id = split, split_id
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        filenames = ['out1_node_feature_label.txt', 'out1_graph_edges.txt'
                     ] + [f'film_split_0.6_0.2_{i}.npz' for i in range(10)]
        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        return osp.join(self._processed_dir, f"{self._name}.graph")

    def _download(self):
        url = 'https://raw.githubusercontent.com/graphdml-uiuc-jlu/geom-gcn/master'

        for raw_file_path in self.raw_file_paths[:2]:
            raw_file_name = osp.basename(raw_file_path)
            file_url = f'{url}/new_data/film/{raw_file_name}'
            print(file_url)
            download_to(file_url, raw_file_path)

        for raw_file_path in self.raw_file_paths[2:]:
            raw_file_name = osp.basename(raw_file_path)
            file_url = f'{url}/splits/{raw_file_name}'
            print(file_url)
            download_to(file_url, raw_file_path)

    def _process(self):
        with open(self.raw_file_paths[0], 'r') as f:
            data = [x.split('\t') for x in f.read().split('\n')[1:-1]]

            rows, cols = [], []
            for n_id, col, _ in data:
                col = [int(x) for x in col.split(',')]
                rows += [int(n_id)] * len(col)
                cols += col
            x = SparseTensor(row=torch.tensor(rows), col=torch.tensor(cols))
            x = x.to_dense()

            labels = torch.empty(len(data), dtype=torch.long)
            for n_id, _, label in data:
                labels[int(n_id)] = int(label)

        features = x.numpy()
        num_node = features.shape[0]
        node_type = "actor"

        with open(self.raw_file_paths[1], 'r') as f:
            data = f.read().split('\n')[1:-1]
            data = [[int(v) for v in r.split('\t')] for r in data]
            edge_index = torch.tensor(data, dtype=torch.long).t().contiguous()
            edge_index, _ = coalesce(edge_index, None, x.size(0), x.size(0))

        row, col = edge_index[0], edge_index[1]
        edge_weight = torch.ones(len(row))
        edge_type = "actor__to__actor"

        g = Graph(row, col, edge_weight, num_node,
                  node_type, edge_type, x=features, y=labels)

        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            train_masks, val_masks, test_masks = [], [], []
            for f in self.raw_file_paths[2:]:
                tmp = np.load(f)
                train_masks += [torch.from_numpy(tmp['train_mask']).to(torch.bool)]
                val_masks += [torch.from_numpy(tmp['val_mask']).to(torch.bool)]
                test_masks += [torch.from_numpy(tmp['test_mask']).to(torch.bool)]
            train_mask = train_masks[self._split_id]
            val_mask = val_masks[self._split_id]
            test_mask = test_masks[self._split_id]

            train_idx = torch.nonzero(train_mask == 1).reshape(-1)
            val_idx = torch.nonzero(val_mask == 1).reshape(-1)
            test_idx = torch.nonzero(test_mask == 1).reshape(-1)

        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import torch
import torch.nn.functional as F
from torch.nn import ModuleList

from sgl.models.simple_models import MultiLayerPerceptron
from sgl.operators.base_op import MessageOp


class ProjectedConcatMessageOp(MessageOp):
    def __init__(self, start, end, feat_dim, hidden_dim, num_layers):
        super(ProjectedConcatMessageOp, self).__init__(start, end)
        self._aggr_type = "proj_concat"

        self.__learnable_weight = ModuleList()
        for _ in range(end - start):
            self.__learnable_weight.append(MultiLayerPerceptron(
                feat_dim, hidden_dim, num_layers, hidden_dim))

    def _combine(self, feat_list):
        adopted_feat_list = feat_list[self._start:self._end]

        concat_feat = self.__learnable_weight[0](adopted_feat_list[0])
        for i in range(1, self._end - self._start):
            transformed_feat = F.relu(
                self.__learnable_weight[i](adopted_feat_list[i]))
            concat_feat = torch.hstack((concat_feat, transformed_feat))

        return concat_feat



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import LogisticRegression
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import LastMessageOp


class SGC(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(SGC, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = LastMessageOp()
        self._base_model = LogisticRegression(feat_dim, output_dim)



from sgl.operators.base_op import MessageOp


class MeanMessageOp(MessageOp):
    def __init__(self, start, end):
        super(MeanMessageOp, self).__init__(start, end)
        self._aggr_type = "mean"

    def _combine(self, feat_list):
        return sum(feat_list[self._start:self._end]) / (self._end - self._start)



from sgl.models.base_model import FastBaseHeteroSGAPModel
from sgl.models.simple_models import MultiLayerPerceptron, FastOneDimConvolution
from sgl.operators.graph_op import LaplacianGraphOp


class Fast_NARS_SGC_WithLearnableWeights(FastBaseHeteroSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers, random_subgraph_num):
        super(Fast_NARS_SGC_WithLearnableWeights, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)

        self._aggregator = FastOneDimConvolution(
            random_subgraph_num, prop_steps + 1)
        self._base_model = MultiLayerPerceptron(
            feat_dim, hidden_dim, num_layers, output_dim)

    @property
    def subgraph_weight(self):
        return self._aggregator.subgraph_weight



import os.path as osp
import torch


def file_exist(filepaths):
    if isinstance(filepaths, list):
        for filepath in filepaths:
            if not osp.exists(filepath):
                return False
        return True
    else:
        if osp.exists(filepaths):
            return True
        else:
            return False


def to_undirected(edge_index):
    row, col = edge_index
    new_row = torch.hstack((row, col))
    new_col = torch.hstack((col, row))
    new_edge_index = torch.stack((new_row, new_col), dim=0)

    return new_edge_index



import numpy as np
import os.path as osp
import pickle as pkl
import torch
from ogb.nodeproppred import PygNodePropPredDataset

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, to_undirected


class Ogbn(NodeDataset):
    def __init__(self, name="arxiv", root="./", split="official"):
        if name not in ["arxiv", "products"]:
            raise ValueError("Dataset name not found!")
        super(Ogbn, self).__init__(root + "ogbn/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filepath = "ogbn_" + self._name + "/raw/geometric_data_processed.pt"
        return osp.join(self._raw_dir, filepath)

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)

    def _process(self):
        dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)

        data = dataset[0]
        features, labels = data.x.numpy().astype(np.float32), data.y.to(torch.long).squeeze(1)
        num_node = data.num_nodes
        if self._name == "products":
            node_type = "product"
        else:
            node_type = "paper"

        undi_edge_index = to_undirected(data.edge_index)
        row, col = undi_edge_index
        edge_weight = torch.ones(len(row))
        if self._name == "products":
            edge_type = "product__to__product"
        else:
            edge_type = "paper__to__paper"

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)
            split_idx = dataset.get_idx_split()
            train_idx, val_idx, test_idx = split_idx['train'], split_idx['valid'], split_idx['test']
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import torch

from sgl.operators.base_op import MessageOp


class ConcatMessageOp(MessageOp):
    def __init__(self, start, end):
        super(ConcatMessageOp, self).__init__(start, end)
        self._aggr_type = "concat"

    def _combine(self, feat_list):
        return torch.hstack(feat_list[self._start:self._end])



import torch
import torch.nn.functional as F
from torch import nn
from torch.nn import Parameter, Linear

from sgl.operators.base_op import MessageOp
from sgl.operators.utils import one_dim_weighted_add, two_dim_weighted_add


class LearnableWeightedMessageOp(MessageOp):

    # 'simple' needs one additional parameter 'prop_steps';
    # 'simple_weighted' allows negative weights, all else being the same as 'simple';
    # 'gate' needs one additional parameter 'feat_dim';
    # 'ori_ref' needs one additional parameter 'feat_dim';
    # 'jk' needs two additional parameter 'prop_steps' and 'feat_dim'
    def __init__(self, start, end, combination_type, *args):
        super(LearnableWeightedMessageOp, self).__init__(start, end)
        self._aggr_type = "learnable_weighted"

        if combination_type not in ["simple", "simple_allow_neg", "gate", "ori_ref", "jk"]:
            raise ValueError(
                "Invalid weighted combination type! Type must be 'simple', 'simple_allow_neg', 'gate', 'ori_ref' or 'jk'.")
        self.__combination_type = combination_type

        self.__learnable_weight = None
        if combination_type == "simple" or combination_type == "simple_allow_neg":
            if len(args) != 1:
                raise ValueError(
                    "Invalid parameter numbers for the simple learnable weighted aggregator!")
            prop_steps = args[0]
            # a 2d tensor is required to use xavier_uniform_.
            tmp_2d_tensor = torch.FloatTensor(1, prop_steps + 1)
            nn.init.xavier_normal_(tmp_2d_tensor)
            self.__learnable_weight = Parameter(tmp_2d_tensor.view(-1))

        elif combination_type == "gate":
            if len(args) != 1:
                raise ValueError(
                    "Invalid parameter numbers for the gate learnable weighted aggregator!")
            feat_dim = args[0]
            self.__learnable_weight = Linear(feat_dim, 1)

        elif combination_type == "ori_ref":
            if len(args) != 1:
                raise ValueError(
                    "Invalid parameter numbers for the ori_ref learnable weighted aggregator!")
            feat_dim = args[0]
            self.__learnable_weight = Linear(feat_dim + feat_dim, 1)

        elif combination_type == "jk":
            if len(args) != 2:
                raise ValueError(
                    "Invalid parameter numbers for the jk learnable weighted aggregator!")
            prop_steps, feat_dim = args[0], args[1]
            self.__learnable_weight = Linear(
                feat_dim + (prop_steps + 1) * feat_dim, 1)

    def _combine(self, feat_list):
        weight_list = None
        if self.__combination_type == "simple":
            weight_list = F.softmax(torch.sigmoid(
                self.__learnable_weight[self._start:self._end]), dim=0)

        elif self.__combination_type == "simple_allow_neg":
            weight_list = self.__learnable_weight[self._start:self._end]

        elif self.__combination_type == "gate":
            adopted_feat_list = torch.vstack(feat_list[self._start:self._end])
            weight_list = F.softmax(
                torch.sigmoid(self.__learnable_weight(adopted_feat_list).view(self._end - self._start, -1).T), dim=1)

        elif self.__combination_type == "ori_ref":
            reference_feat = feat_list[0].repeat(self._end - self._start, 1)
            adopted_feat_list = torch.hstack(
                (reference_feat, torch.vstack(feat_list[self._start:self._end])))
            weight_list = F.softmax(
                torch.sigmoid(self.__learnable_weight(adopted_feat_list).view(-1, self._end - self._start)), dim=1)

        elif self.__combination_type == "jk":
            reference_feat = torch.hstack(feat_list).repeat(
                self._end - self._start, 1)
            adopted_feat_list = torch.hstack(
                (reference_feat, torch.vstack(feat_list[self._start:self._end])))
            weight_list = F.softmax(
                torch.sigmoid(self.__learnable_weight(adopted_feat_list).view(-1, self._end - self._start)), dim=1)

        else:
            raise NotImplementedError

        weighted_feat = None
        if self.__combination_type == "simple" or self.__combination_type == "simple_allow_neg":
            weighted_feat = one_dim_weighted_add(
                feat_list[self._start:self._end], weight_list=weight_list)
        elif self.__combination_type in ["gate", "ori_ref", "jk"]:
            weighted_feat = two_dim_weighted_add(
                feat_list[self._start:self._end], weight_list=weight_list)
        else:
            raise NotImplementedError

        return weighted_feat



import torch
import torch.nn.functional as F

from sgl.operators.base_op import MessageOp

class OverSmoothDistanceWeightedOp(MessageOp):
    def __init__(self):
        super(OverSmoothDistanceWeightedOp, self).__init__()
        self._aggr_type = 'over_smooth_dis_weighted'

    def _combine(self, feat_list): 
        weight_list = []
        features = feat_list[0]
        norm_fea = torch.norm(features, 2, 1).add(1e-10)
        for fea in feat_list:
            norm_cur = torch.norm(fea, 2, 1).add(1e-10)
            tmp = torch.div((features * fea).sum(1), norm_cur)
            tmp = torch.div(tmp, norm_fea)

            weight_list.append(tmp.unsqueeze(-1))

        weight = F.softmax(torch.cat(weight_list, dim=1), dim=1)

        hops = len(feat_list)
        num_nodes = features.shape[0]
        output = []
        for i in range(num_nodes):
            fea = 0.
            for j in range(hops):
                fea += (weight[i][j]*feat_list[j][i]).unsqueeze(0)
            output.append(fea)
        output = torch.cat(output, dim=0)
        return output 



import scipy.sparse as sp
import numpy as np
import torch
import torch.nn.functional as F
import math

def LogeCrossEntropy(pred, target, epsilon= 1.0 - math.log(2)):
    loss = F.cross_entropy(pred, target)
    loss = torch.log(epsilon + loss) - math.log(epsilon)
    return torch.mean(loss)

def LogeBinaryCrossEntropyWithLogits(pred, target, epsilon= 1.0 - math.log(2)):
    loss = F.binary_cross_entropy_with_logits(pred, target)
    loss = torch.log(epsilon + loss) - math.log(epsilon)
    return loss

def adj_to_symmetric_norm(adj, r):
    adj = adj + sp.eye(adj.shape[0])
    degrees = np.array(adj.sum(1))
    r_inv_sqrt_left = np.power(degrees, r - 1).flatten()
    r_inv_sqrt_left[np.isinf(r_inv_sqrt_left)] = 0.
    r_mat_inv_sqrt_left = sp.diags(r_inv_sqrt_left)

    r_inv_sqrt_right = np.power(degrees, -r).flatten()
    r_inv_sqrt_right[np.isinf(r_inv_sqrt_right)] = 0.
    r_mat_inv_sqrt_right = sp.diags(r_inv_sqrt_right)

    adj_normalized = adj.dot(r_mat_inv_sqrt_left).transpose().dot(r_mat_inv_sqrt_right)
    return adj_normalized
    
def sparse_mx_to_torch_sparse_tensor(sparse_mx):
    """Convert a scipy sparse matrix to a torch sparse tensor."""
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse.FloatTensor(indices, values, shape)

@torch.no_grad()
def label_propagation(labels, adj, num_layers, alpha, post_process = lambda x: x.clamp_(0., 1.), mask=None):
    if labels.dtype == torch.long:
        labels = F.one_hot(labels.reshape(-1)).to(torch.float)

    out = labels.clone()
    if mask is not None:
        out = torch.zeros_like(labels)
        out[mask] = labels[mask]

    adj_tensor = sparse_mx_to_torch_sparse_tensor(adj)
    
    # record H_0
    res = (1 - alpha) * out
    for _ in range(num_layers):
        out = alpha * torch.spmm(adj_tensor, out) + res
        out = post_process(out)
    
    return out



import os.path as osp
import pickle as pkl
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to, read_npz, random_split_dataset


class Amazon(NodeDataset):
    def __init__(self, name="photo", root="./", split="random"):
        if name not in ['computers', 'photo']:
            raise ValueError("Dataset name not supported!")
        super(Amazon, self).__init__(root + "amazon/", name)
        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        return osp.join(self._raw_dir, f'amazon_electronics_{self.name.lower()}.npz')

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        url = "https://github.com/shchur/gnn-benchmark/raw/master/data/npz"
        file_url = url + '/' + osp.basename(self.raw_file_paths)
        print(file_url)
        download_to(file_url, self.raw_file_paths)

    def _process(self):
        x, edge_index, y = read_npz(self.raw_file_paths)
        num_node = x.shape[0]
        row, col = edge_index
        edge_weight = torch.ones(size=(len(row),))
        node_type = "product"
        edge_type = "product__to__product"
        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=x.numpy(), y=y)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "random":
            train_idx, val_idx, test_idx = random_split_dataset(self.num_node)
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import argparse
import numpy as np
from sgl.search.auto_search_dist import SearchManagerDist
from sgl.search.search_models_dist import SearchModelDist
from openbox.utils.config_space import ConfigurationSpace, UniformIntegerHyperparameter

class ConfigManagerDist():
    def __init__(self, arch, prop_steps=[1,10], prop_types=[0,1], mesg_types=[0,8], num_layers=[1,10], post_steps=[1,10], post_types=[0,1], pmsg_types=[0,5]):
        super(ConfigManagerDist, self).__init__()

        self.__initial_arch = arch
        self.__config_space = ConfigurationSpace()
        self.__prop_steps = UniformIntegerHyperparameter("prop_steps", prop_steps[0], prop_steps[1])
        self.__prop_types = UniformIntegerHyperparameter("prop_types", prop_types[0], prop_types[1])
        self.__mesg_types = UniformIntegerHyperparameter("mesg_types", mesg_types[0], mesg_types[1])
        self.__num_layers = UniformIntegerHyperparameter("num_layers", num_layers[0], num_layers[1])
        self.__post_steps = UniformIntegerHyperparameter("post_steps", post_steps[0], post_steps[1])
        self.__post_types = UniformIntegerHyperparameter("post_types", post_types[0], post_types[1])
        self.__pmsg_types = UniformIntegerHyperparameter("pmsg_types", pmsg_types[0], pmsg_types[1])
        self.__config_space.add_hyperparameters([self.__prop_steps, self.__prop_types, self.__mesg_types, self.__num_layers, self.__post_steps, self.__post_types, self.__pmsg_types])
    
    def _setParameters(self, dataset, device, hiddim, epochs, lr, wd):
        self.__dataset = dataset
        self.__device = device
        self.__hiddim = hiddim
        self.__epochs = epochs
        self.__lr = lr
        self.__wd = wd

        parser = argparse.ArgumentParser()
        parser.add_argument('-n', '--nodes', default=1, type=int, metavar='N',
                            help='number of data loading workers (default: 4)')
        parser.add_argument('-g', '--gpus', default=2, type=int,
                            help='number of gpus per node')
        parser.add_argument('-nr', '--nr', default=0, type=int,
                            help='ranking within the nodes')
        parser.add_argument('--lr', default=1e-2, type=float, metavar='LR',
                            help='learning rate')
        parser.add_argument('--wd', default=5e-4, type=float, metavar='WD',
                            help='weight decay')                        
        parser.add_argument('--epochs', default=200, type=int, metavar='N',
                            help='number of total epochs to run')
        parser.add_argument('--hidden', default=128, type=int, metavar='N',
                            help='hidden_dims')
        parser.add_argument('--batch', default=128, type=int, metavar='N',
                            help='batch_size')
        args = parser.parse_args()


    def _configSpace(self):
        return self.__config_space

    def _configTarget(self, arch):
        model = SearchModelDist(arch, self.__dataset.num_features, int(self.__dataset.num_classes), self.__hiddim)
        acc_res, time_res = SearchManagerDist(self.__dataset, model)._execute(args)
        result = dict()
        result['objs'] = np.stack([-acc_res, time_res], axis=-1)
        return result

    def _configFunction(self, config_space):
        self.__initial_arch[0] = config_space['prop_steps']
        self.__initial_arch[1] = config_space['prop_types']
        self.__initial_arch[2] = config_space['mesg_types']
        self.__initial_arch[3] = config_space['num_layers']
        self.__initial_arch[4] = config_space['post_steps']
        self.__initial_arch[5] = config_space['post_types']
        self.__initial_arch[6] = config_space['pmsg_types']
        result = self._configTarget(self.__initial_arch)
        return result



import numpy as np
import random
import torch


def accuracy(output, labels):
    pred = output.max(1)[1].type_as(labels)
    correct = pred.eq(labels).double()
    correct = correct.sum()
    return (correct / len(labels)).item()


def set_seed(seed):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def evaluate(model, val_idx, test_idx, labels, device):
    model.eval()
    val_output = model.model_forward(val_idx, device)
    test_output = model.model_forward(test_idx, device)

    acc_val = accuracy(val_output, labels[val_idx])
    acc_test = accuracy(test_output, labels[test_idx])
    return acc_val, acc_test


def mini_batch_evaluate(model, val_idx, val_loader, test_idx, test_loader, labels, device):
    model.eval()
    correct_num_val, correct_num_test = 0, 0
    for batch in val_loader:
        val_output = model.model_forward(batch, device)
        pred = val_output.max(1)[1].type_as(labels)
        correct_num_val += pred.eq(labels[batch]).double().sum()
    acc_val = correct_num_val / len(val_idx)

    for batch in test_loader:
        test_output = model.model_forward(batch, device)
        pred = test_output.max(1)[1].type_as(labels)
        correct_num_test += pred.eq(labels[batch]).double().sum()
    acc_test = correct_num_test / len(test_idx)

    return acc_val.item(), acc_test.item()


def train(model, train_idx, labels, device, optimizer, loss_fn):
    model.train()
    optimizer.zero_grad()

    train_output = model.model_forward(train_idx, device)
    loss_train = loss_fn(train_output, labels[train_idx])
    acc_train = accuracy(train_output, labels[train_idx])
    loss_train.backward()
    optimizer.step()

    return loss_train.item(), acc_train


def mini_batch_train(model, train_idx, train_loader, labels, device, optimizer, loss_fn):
    model.train()
    correct_num = 0
    loss_train_sum = 0.
    for batch in train_loader:
        train_output = model.model_forward(batch, device)
        loss_train = loss_fn(train_output, labels[batch])

        pred = train_output.max(1)[1].type_as(labels)
        correct_num += pred.eq(labels[batch]).double().sum()
        loss_train_sum += loss_train.item()

        optimizer.zero_grad()
        loss_train.backward()
        optimizer.step()

    loss_train = loss_train_sum / len(train_loader)
    acc_train = correct_num / len(train_idx)

    return loss_train, acc_train.item()



import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import torch
from ogb.nodeproppred import PygNodePropPredDataset

from sgl.data.base_data import HeteroGraph
from sgl.data.base_dataset import HeteroNodeDataset
from sgl.dataset.utils import pkl_read_file, to_undirected


class OgbnMag(HeteroNodeDataset):
    TYPE_OF_NODE_TO_PREDICT = 'paper'

    def __init__(self, name="mag", root=".", split="official"):
        if name not in ["mag"]:
            raise ValueError("Dataset name not found!")
        super(OgbnMag, self).__init__(root + "ogbn/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filepath = "ogbn_" + self._name + "/raw/geometric_data_processed.pt"
        return osp.join(self._raw_dir, filepath)

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    @property
    def EDGE_TYPES(self):
        return self._data.edge_types

    def _download(self):
        dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)

    def _process(self):
        dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)

        data = dataset[0]
        node_types = list(data.num_nodes_dict.keys())
        edge_types = list(data.edge_index_dict.keys())

        # generate feature for nodes without features via averaging
        data.x_dict['paper'] = data.x_dict['paper'].numpy()

        row_author_temp = data.edge_index_dict[('author', 'writes', 'paper')][0, :].numpy()
        col_author_temp = data.edge_index_dict[('author', 'writes', 'paper')][1, :].numpy()
        edge_weight_author_temp = torch.ones(len(row_author_temp)).numpy()
        adj_author_temp = sp.csr_matrix((edge_weight_author_temp, (row_author_temp, col_author_temp)),
                                        shape=(data.num_nodes_dict['author'], data.num_nodes_dict['paper']))
        data.x_dict['author'] = adj_author_temp.dot(data.x_dict['paper'])

        row_inst_temp = data.edge_index_dict[('author', 'affiliated_with', 'institution')][1, :].numpy()
        col_inst_temp = data.edge_index_dict[('author', 'affiliated_with', 'institution')][0, :].numpy()
        edge_weight_inst_temp = torch.ones(len(row_inst_temp)).numpy()
        adj_inst_temp = sp.csr_matrix((edge_weight_inst_temp, (row_inst_temp, col_inst_temp)),
                                      shape=(data.num_nodes_dict['institution'], data.num_nodes_dict['author']))
        data.x_dict['institution'] = adj_inst_temp.dot(data.x_dict['author'])

        row_field_temp = data.edge_index_dict[('paper', 'has_topic', 'field_of_study')][1, :].numpy()
        col_field_temp = data.edge_index_dict[('paper', 'has_topic', 'field_of_study')][0, :].numpy()
        edge_weight_field_temp = torch.ones(len(row_field_temp)).numpy()
        adj_field_temp = sp.csr_matrix((edge_weight_field_temp, (row_field_temp, col_field_temp)),
                                       shape=(data.num_nodes_dict['field_of_study'], data.num_nodes_dict['paper']))
        data.x_dict['field_of_study'] = adj_field_temp.dot(data.x_dict['paper'])

        current_nodes = [0]
        current_nodes_dict = {}

        # the order of node types is important, which decides the node id
        num_node_dict, node_id_dict, x_dict, y_dict = {}, {}, {}, {}
        for i, node_type in enumerate(node_types):
            num_nodes_temp = data.num_nodes_dict[node_type]
            x_temp = data.x_dict.get(node_type, None)
            y_temp = data.y_dict.get(node_type, None)

            node_ids_temp = list(range(current_nodes[i], current_nodes[i] + data.num_nodes_dict[node_type]))
            current_nodes.append(current_nodes[-1] + data.num_nodes_dict[node_type])
            current_nodes_dict[node_type] = current_nodes[i]

            num_node_dict[node_type] = num_nodes_temp
            node_id_dict[node_type] = node_ids_temp
            x_dict[node_type] = x_temp
            y_dict[node_type] = y_temp

        edge_types_found = []
        row_dict, col_dict, edge_weight_dict = {}, {}, {}
        for i, edge_type in enumerate(edge_types):
            row_type = edge_type[0]
            col_type = edge_type[2]

            edge_type_used = "__".join([row_type, "to", col_type])
            edge_types_found.append(edge_type_used)

            row_temp = data.edge_index_dict[edge_type][0, :] + current_nodes_dict[row_type]
            col_temp = data.edge_index_dict[edge_type][1, :] + current_nodes_dict[col_type]

            if row_type != col_type:
                edge_weight_temp = torch.ones(len(row_temp))

                row_dict[edge_type_used] = row_temp
                col_dict[edge_type_used] = col_temp
                edge_weight_dict[edge_type_used] = edge_weight_temp

                reverse_edge_type_used = "__".join([col_type, "to", row_type])
                edge_types_found.append(reverse_edge_type_used)

                row_dict[reverse_edge_type_used] = col_temp
                col_dict[reverse_edge_type_used] = row_temp
                edge_weight_dict[reverse_edge_type_used] = edge_weight_temp

            else:
                row_temp, col_temp = to_undirected((row_temp, col_temp))
                edge_weight_temp = torch.ones(len(row_temp))

                row_dict[edge_type_used] = row_temp
                col_dict[edge_type_used] = col_temp
                edge_weight_dict[edge_type_used] = edge_weight_temp

        g = HeteroGraph(row_dict, col_dict, edge_weight_dict, num_node_dict, node_types, edge_types_found, node_id_dict,
                        x_dict, y_dict)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            dataset = PygNodePropPredDataset("ogbn-" + self._name, self._raw_dir)
            split_idx = dataset.get_idx_split()
            train_idx, val_idx, test_idx = split_idx['train']['paper'], split_idx['valid']['paper'], split_idx['test'][
                'paper']
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import torch
import torch.nn.functional as F
from torch.nn import Linear

from sgl.operators.base_op import MessageOp


class IterateLearnableWeightedMessageOp(MessageOp):

    # 'recursive' needs one additional parameter 'feat_dim'
    def __init__(self, start, end, combination_type, *args):
        super(IterateLearnableWeightedMessageOp, self).__init__(start, end)
        self._aggr_type = "iterate_learnable_weighted"

        if combination_type not in ["recursive"]:
            raise ValueError(
                "Invalid weighted combination type! Type must be 'recursive'.")
        self.__combination_type = combination_type

        self.__learnable_weight = None
        if combination_type == "recursive":
            if len(args) != 1:
                raise ValueError(
                    "Invalid parameter numbers for the recursive iterate weighted aggregator!")
            feat_dim = args[0]
            self.__learnable_weight = Linear(feat_dim + feat_dim, 1)

    def _combine(self, feat_list):
        weight_list = None
        if self.__combination_type == "recursive":
            weighted_feat = feat_list[self._start]
            for i in range(self._start, self._end):
                weights = torch.sigmoid(self.__learnable_weight(
                    torch.hstack((feat_list[i], weighted_feat))))
                if i == self._start:
                    weight_list = weights
                else:
                    weight_list = torch.hstack((weight_list, weights))
                weight_list = F.softmax(weight_list, dim=1)

                weighted_feat = torch.mul(
                    feat_list[self._start], weight_list[:, 0].view(-1, 1))
                for j in range(1, i + 1):
                    weighted_feat = weighted_feat + \
                                    torch.mul(feat_list[self._start + j],
                                              weight_list[:, j].view(-1, 1))

        else:
            raise NotImplementedError

        return weighted_feat



import time
import torch
import torch.nn as nn
from torch.optim import Adam

from sgl.search.base_search import BaseSearch
from sgl.search.utils import accuracy, set_seed


class SearchManager(BaseSearch):
    def __init__(self, dataset, model, lr, weight_decay, epochs, device, loss_fn=nn.CrossEntropyLoss(), seed=42):
        super(SearchManager, self).__init__()

        self.__dataset = dataset
        self.__model = model
        self.__optimizer = Adam(model.parameters(), lr=lr, weight_decay=weight_decay)
        self.__epochs = epochs
        self.__loss_fn = loss_fn
        self.__device = device
        self.__seed = seed

    def _execute(self):
        set_seed(self.__seed)
        t_pre_start = time.time()
        self.__model.preprocess(self.__dataset.adj, self.__dataset.x)
        t_pre_end = time.time()
        self.__model = self.__model.to(self.__device)
        self.__dataset.y = self.__dataset.y.to(self.__device)
        time_preprocess = t_pre_end - t_pre_start

        t_total = time.time()
        best_val = 0.
        best_test = 0.
        for i in range(10):
            for epoch in range(self.__epochs):
                t = time.time()
                loss_train, acc_train = self._train()
                acc_val, acc_test = self._evaluate()
                print('Epoch: {:03d}'.format(epoch + 1),
                      'loss_train: {:.4f}'.format(loss_train),
                      'acc_train: {:.4f}'.format(acc_train),
                      'acc_val: {:.4f}'.format(acc_val),
                      'acc_test: {:.4f}'.format(acc_test),
                      'time: {:.4f}s'.format(time.time() - t))
                if acc_val > best_val:
                    best_val = acc_val
                    best_test = acc_test
                    torch.save(self.__model, './best.pt')

        acc_val, acc_test, time_forward = self._postprocess()
        if acc_val > best_val:
            best_val = acc_val
            best_test = acc_test
        total_time = time_preprocess + time_forward

        print("Optimization Finished!")
        print("Total time elapsed: {:.4f}s".format(time.time() - t_total))
        print(f'Best val: {best_val:.4f}, best test: {best_test:.4f}')
        return best_test, total_time

    def _postprocess(self):
        model = torch.load('./best.pt')
        model.eval()
        t_forward_start = time.time()
        output = model.model_forward(range(self.__dataset.num_node), self.__device)
        final_output = model.postprocess(self.__dataset.adj, output)
        t_forward_end = time.time()
        time_forward = t_forward_end - t_forward_start

        acc_val = accuracy(final_output[self.__dataset.val_idx], self.__dataset.y[self.__dataset.val_idx])
        acc_test = accuracy(final_output[self.__dataset.test_idx], self.__dataset.y[self.__dataset.test_idx])
        return acc_val, acc_test, time_forward

    def _evaluate(self):
        self.__model.eval()
        val_output = self.__model.model_forward(self.__dataset.val_idx, self.__device)
        test_output = self.__model.model_forward(self.__dataset.test_idx, self.__device)

        acc_val = accuracy(val_output, self.__dataset.y[self.__dataset.val_idx])
        acc_test = accuracy(test_output, self.__dataset.y[self.__dataset.test_idx])
        return acc_val, acc_test

    def _train(self):
        self.__model.train()
        self.__optimizer.zero_grad()

        train_output = self.__model.model_forward(self.__dataset.train_idx, self.__device)
        loss_train = self.__loss_fn(train_output, self.__dataset.y[self.__dataset.train_idx])
        acc_train = accuracy(train_output, self.__dataset.y[self.__dataset.train_idx])
        loss_train.backward()
        self.__optimizer.step()

        return loss_train.item(), acc_train



import torch
import torch.nn as nn


class OneDimConvolution(nn.Module):
    def __init__(self, num_subgraphs, prop_steps, feat_dim):
        super(OneDimConvolution, self).__init__()
        self.__hop_num = prop_steps

        self.__learnable_weight = nn.ParameterList()
        for _ in range(prop_steps):
            self.__learnable_weight.append(nn.Parameter(
                torch.FloatTensor(feat_dim, num_subgraphs)))

        self.reset_parameters()

    def reset_parameters(self):
        for weight in self.__learnable_weight:
            nn.init.xavier_uniform_(weight)

    # feat_list_list = hop_num * feat_list = hop_num * (subgraph_num * feat)
    def forward(self, feat_list_list):
        aggregated_feat_list = []
        for i in range(self.__hop_num):
            adopted_feat = torch.stack(feat_list_list[i], dim=2)
            intermediate_feat = (
                    adopted_feat * (self.__learnable_weight[i].unsqueeze(dim=0))).mean(dim=2)

            aggregated_feat_list.append(intermediate_feat)

        return aggregated_feat_list


class OneDimConvolutionWeightSharedAcrossFeatures(nn.Module):
    def __init__(self, num_subgraphs, prop_steps):
        super(OneDimConvolutionWeightSharedAcrossFeatures, self).__init__()
        self.__hop_num = prop_steps

        self.__learnable_weight = nn.ParameterList()
        for _ in range(prop_steps):
            # To help xvarient_uniform_ calculate fan in and fan out, "1" should be kept here.
            self.__learnable_weight.append(nn.Parameter(
                torch.FloatTensor(1, num_subgraphs)))

        self.reset_parameters()

    def reset_parameters(self):
        for weight in self.__learnable_weight:
            nn.init.xavier_uniform_(weight)

    # feat_list_list = hop_num * feat_list = hop_num * (subgraph_num * feat)
    def forward(self, feat_list_list):
        aggregated_feat_list = []
        for i in range(self.__hop_num):
            adopted_feat = torch.stack(feat_list_list[i], dim=2)
            intermediate_feat = (
                    adopted_feat * (self.__learnable_weight[i])).mean(dim=2)

            aggregated_feat_list.append(intermediate_feat)

        return aggregated_feat_list


class FastOneDimConvolution(nn.Module):
    def __init__(self, num_subgraphs, prop_steps):
        super(FastOneDimConvolution, self).__init__()

        self.__num_subgraphs = num_subgraphs
        self.__prop_steps = prop_steps

        # How to initialize the weight is extremely important.
        # Pure xavier will lead to extremely unstable accuracy.
        # Initialized with ones will not perform as good as this one.        
        self.__learnable_weight = nn.Parameter(
            torch.ones(num_subgraphs * prop_steps, 1))

    # feat_list_list: 3-d tensor (num_node, feat_dim, num_subgraphs * prop_steps)
    def forward(self, feat_list_list):
        return (feat_list_list @ self.__learnable_weight).squeeze(dim=2)

    @property
    def subgraph_weight(self):
        return self.__learnable_weight.view(
            self.__num_subgraphs, self.__prop_steps).sum(dim=1)

class IdenticalMapping(nn.Module):
    def __init__(self) -> None:
        super(IdenticalMapping, self).__init__()

    def forward(self, feature):
        return feature
        
class LogisticRegression(nn.Module):
    def __init__(self, feat_dim, output_dim):
        super(LogisticRegression, self).__init__()
        self.__fc = nn.Linear(feat_dim, output_dim)

    def forward(self, feature):
        output = self.__fc(feature)
        return output


class MultiLayerPerceptron(nn.Module):
    def __init__(self, feat_dim, hidden_dim, num_layers, output_dim, dropout=0.5, bn=False):
        super(MultiLayerPerceptron, self).__init__()
        if num_layers < 2:
            raise ValueError("MLP must have at least two layers!")
        self.__num_layers = num_layers

        self.__fcs = nn.ModuleList()
        self.__fcs.append(nn.Linear(feat_dim, hidden_dim))
        for _ in range(num_layers - 2):
            self.__fcs.append(nn.Linear(hidden_dim, hidden_dim))
        self.__fcs.append(nn.Linear(hidden_dim, output_dim))

        self.__bn = bn
        if self.__bn is True:
            self.__bns = nn.ModuleList()
            for _ in range(num_layers - 1):
                self.__bns.append(nn.BatchNorm1d(hidden_dim))

        self.__dropout = nn.Dropout(dropout)
        self.__prelu = nn.PReLU()
        self.reset_parameters()

    def reset_parameters(self):
        gain = nn.init.calculate_gain("relu")
        for fc in self.__fcs:
            nn.init.xavier_uniform_(fc.weight, gain=gain)
            nn.init.zeros_(fc.bias)

    def forward(self, feature):
        for i in range(self.__num_layers - 1):
            feature = self.__fcs[i](feature)
            if self.__bn is True:
                feature = self.__bns[i](feature)
            feature = self.__prelu(feature)
            feature = self.__dropout(feature)

        output = self.__fcs[-1](feature)
        return output

class ResMultiLayerPerceptron(nn.Module):
    def __init__(self, feat_dim, hidden_dim, num_layers, output_dim, dropout=0.8, bn=False):
        super(ResMultiLayerPerceptron, self).__init__()
        if num_layers < 2:
            raise ValueError("ResMLP must have at least two layers!")
        self.__num_layers = num_layers

        self.__fcs = nn.ModuleList()
        self.__fcs.append(nn.Linear(feat_dim, hidden_dim))
        for _ in range(num_layers - 2):
            self.__fcs.append(nn.Linear(hidden_dim, hidden_dim))
        self.__fcs.append(nn.Linear(hidden_dim, output_dim))

        self.__bn = bn
        if self.__bn is True:
            self.__bns = nn.ModuleList()
            for _ in range(num_layers - 1):
                self.__bns.append(nn.BatchNorm1d(hidden_dim))

        self.__dropout = nn.Dropout(dropout)
        self.__relu = nn.ReLU()

    def forward(self, feature):
        feature = self.__dropout(feature)
        feature = self.__fcs[0](feature)
        if self.__bn is True:
            feature = self.__bns[0](feature)
        feature = self.__relu(feature)
        residual = feature

        for i in range(1, self.__num_layers - 1):
            feature = self.__dropout(feature)
            feature = self.__fcs[i](feature)
            if self.__bn is True:
                feature = self.__bns[i](feature)
            feature_ = self.__relu(feature)
            feature = feature_ + residual
            residual = feature_

        feature = self.__dropout(feature)
        output = self.__fcs[-1](feature)
        return output



from sgl.models.base_model import BaseHeteroSGAPModel
from sgl.models.simple_models import OneDimConvolution, MultiLayerPerceptron
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import ProjectedConcatMessageOp


class NARS_SIGN(BaseHeteroSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim, hidden_dim, num_layers, random_subgraph_num):
        super(NARS_SIGN, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = ProjectedConcatMessageOp(
            0, prop_steps + 1, feat_dim, hidden_dim, num_layers)

        self._aggregator = OneDimConvolution(
            random_subgraph_num, prop_steps + 1, feat_dim)
        self._base_model = MultiLayerPerceptron(
            hidden_dim * (prop_steps + 1), hidden_dim, num_layers, output_dim)



class BaseSearch:
    def __init__(self):
        pass

    def _execute(self):
        return NotImplementedError

    def _evaluate(self):
        return NotImplementedError

    def _train(self):
        return NotImplementedError



# DO NOT use this one directly. Instead, go to dblp.py
import numpy as np
import os.path as osp
import scipy.sparse as sp
import torch
from itertools import product
from torch_geometric.data import (
    HeteroData,
    InMemoryDataset,
    extract_zip,
)
from typing import Callable, List, Optional


class DBLP(InMemoryDataset):
    r"""A subset of the DBLP computer science bibliography website, as
    collected in the `"MAGNN: Metapath Aggregated Graph Neural Network for
    Heterogeneous Graph Embedding" <https://arxiv.org/abs/2002.01680>`_ paper.
    DBLP is a heterogeneous graph containing four types of entities - authors
    (4,057 nodes), papers (14,328 nodes), terms (7,723 nodes), and conferences
    (20 nodes).
    The authors are divided into four research areas (database, data mining,
    artificial intelligence, information retrieval).
    Each author is described by a bag-of-words representation of their paper
    keywords.

    Args:
        root (string): Root directory where the dataset should be saved.
        transform (callable, optional): A function/transform that takes in an
            :obj:`torch_geometric.data.HeteroData` object and returns a
            transformed version. The data object will be transformed before
            every access. (default: :obj:`None`)
        pre_transform (callable, optional): A function/transform that takes in
            an :obj:`torch_geometric.data.HeteroData` object and returns a
            transformed version. The data object will be transformed before
            being saved to disk. (default: :obj:`None`)
    """

    url = 'https://www.dropbox.com/s/yh4grpeks87ugr2/DBLP_processed.zip?dl=1'

    def __init__(self, root: str, path_of_zip: str, transform: Optional[Callable] = None,
                 pre_transform: Optional[Callable] = None):
        self.__path_of_zip = path_of_zip
        super().__init__(root, transform, pre_transform)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self) -> List[str]:
        return [
            'adjM.npz', 'features_0.npz', 'features_1.npz', 'features_2.npy',
            'labels.npy', 'node_types.npy', 'train_val_test_idx.npz'
        ]

    @property
    def processed_file_names(self) -> str:
        return 'data.pt'

    def download(self):
        path = self.__path_of_zip
        extract_zip(path, self.raw_dir)

    def process(self):
        data = HeteroData()

        node_types = ['author', 'paper', 'term', 'conference']
        for i, node_type in enumerate(node_types[:2]):
            x = sp.load_npz(osp.join(self.raw_dir, f'features_{i}.npz'))
            data[node_type].x = torch.from_numpy(x.todense()).to(torch.float)

        x = np.load(osp.join(self.raw_dir, 'features_2.npy'))
        data['term'].x = torch.from_numpy(x).to(torch.float)

        node_type_idx = np.load(osp.join(self.raw_dir, 'node_types.npy'))
        node_type_idx = torch.from_numpy(node_type_idx).to(torch.long)
        data['conference'].num_nodes = int((node_type_idx == 3).sum())

        y = np.load(osp.join(self.raw_dir, 'labels.npy'))
        data['author'].y = torch.from_numpy(y).to(torch.long)

        split = np.load(osp.join(self.raw_dir, 'train_val_test_idx.npz'))
        for name in ['train', 'val', 'test']:
            idx = split[f'{name}_idx']
            idx = torch.from_numpy(idx).to(torch.long)
            mask = torch.zeros(data['author'].num_nodes, dtype=torch.bool)
            mask[idx] = True
            data['author'][f'{name}_mask'] = mask

        s = {}
        N_a = data['author'].num_nodes
        N_p = data['paper'].num_nodes
        N_t = data['term'].num_nodes
        N_c = data['conference'].num_nodes
        s['author'] = (0, N_a)
        s['paper'] = (N_a, N_a + N_p)
        s['term'] = (N_a + N_p, N_a + N_p + N_t)
        s['conference'] = (N_a + N_p + N_t, N_a + N_p + N_t + N_c)

        A = sp.load_npz(osp.join(self.raw_dir, 'adjM.npz'))
        for src, dst in product(node_types, node_types):
            A_sub = A[s[src][0]:s[src][1], s[dst][0]:s[dst][1]].tocoo()
            if A_sub.nnz > 0:
                row = torch.from_numpy(A_sub.row).to(torch.long)
                col = torch.from_numpy(A_sub.col).to(torch.long)
                data[src, dst].edge_index = torch.stack([row, col], dim=0)

        if self.pre_transform is not None:
            data = self.pre_transform(data)

        torch.save(self.collate([data]), self.processed_paths[0])

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}()'



import setuptools

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

with open("requirements.txt", "r", encoding="utf-8") as require:
    requirements = [r for r in require.read().splitlines() if r != '']

setuptools.setup(
    name="sgl-dair",
    version="0.1.5",
    author="DAIR Lab, PKU",
    description="Graph Neural Network (GNN) toolkit targeting scalable graph learning",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/PKU-DAIR/SGL",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=requirements,
)



# Haven't tested yet
import json
import numpy as np
import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import torch

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


class AmazonProduct(NodeDataset):
    def __init__(self, name="amazonproduct", root="./", split="official"):
        super(AmazonProduct, self).__init__(root + "AmazonProduct", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(
            split)

    @property
    def raw_file_paths(self):
        filenames = ['adj_full.npz', 'feats.npy', 'class_map.json', 'role.json']
        return [osp.join(self._raw_dir, filename) for filename in filenames]

    @property
    def processed_file_paths(self):
        return osp.join(self._processed_dir, f"{self._name}.graph")

    def _download(self):
        url = 'https://docs.google.com/uc?export=download&id={}&confirm=t'

        adj_full_id = '17qhNA8H1IpbkkR-T2BmPQm8QNW5do-aa'
        feats_id = '10SW8lCvAj-kb6ckkfTOC5y0l8XXdtMxj'
        class_map_id = '1LIl4kimLfftj4-7NmValuWyCQE8AaE7P'
        role_id = '1npK9xlmbnjNkV80hK2Q68wTEVOFjnt4K'

        path = osp.join(self._raw_dir, 'adj_full.npz')
        file_url = url.format(adj_full_id)
        print(file_url)
        download_to(file_url, path)

        path = osp.join(self._raw_dir, 'feats.npy')
        file_url = url.format(feats_id)
        print(file_url)
        download_to(file_url, path)

        path = osp.join(self._raw_dir, 'class_map.json')
        file_url = url.format(class_map_id)
        print(file_url)
        download_to(file_url, path)

        path = osp.join(self._raw_dir, 'role.json')
        file_url = url.format(role_id)
        print(file_url)
        download_to(file_url, path)

    def _process(self):
        features = np.load(osp.join(self._raw_dir, 'feats.npy'))
        num_node = features.shape[0]
        node_type = "product"

        f = np.load(osp.join(self._raw_dir, 'adj_full.npz'))
        adj = sp.csr_matrix((f['data'], f['indices'], f['indptr']), f['shape'])
        adj = adj.tocoo()

        row, col, edge_weight = adj.row, adj.col, adj.data
        edge_type = "product__to__product"

        ys = [-1] * num_node
        with open(osp.join(self._raw_dir, 'class_map.json')) as f:
            class_map = json.load(f)
            for key, item in class_map.items():
                ys[int(key)] = item
        labels = torch.LongTensor(ys)

        g = Graph(row, col, edge_weight, num_node,
                  node_type, edge_type, x=features, y=labels)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            with open(osp.join(self._raw_dir, 'role.json')) as f:
                role = json.load(f)

            train_idx, val_idx, test_idx = role['tr'], role['va'], role['te']
        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



import networkx as nx
import os
import os.path as osp
import pickle as pkl
import scipy.sparse as sp
import shutil
import torch
from torch_geometric.data import extract_tar
from torch_geometric.io import read_txt_array

from sgl.data.base_data import Graph
from sgl.data.base_dataset import NodeDataset
from sgl.dataset.utils import pkl_read_file, download_to


class Nell(NodeDataset):
    def __init__(self, name="nell.0.001", root='./', split="official"):
        if name not in ["nell.0.1", "nell.0.01", "nell.0.001"]:
            raise ValueError("Dataset name not supported!")
        super(Nell, self).__init__(root + "Nell/", name)

        self._data = pkl_read_file(self.processed_file_paths)
        self._split = split
        self._train_idx, self._val_idx, self._test_idx = self.__generate_split(split)

    @property
    def raw_file_paths(self):
        filenames = ['x', 'tx', 'allx', 'y', 'ty', 'ally', 'graph', 'test.index']
        return [osp.join(self._raw_dir, f'ind.{self._name}.{filename}') for filename in filenames]

    @property
    def processed_file_paths(self):
        filename = "graph"
        return osp.join(self._processed_dir, "{}.{}".format(self._name, filename))

    def _download(self):
        url = 'http://www.cs.cmu.edu/~zhiliny/data/nell_data.tar.gz'
        path = osp.join(self._raw_dir, "nell_data.tar.gz")
        print(url)
        download_to(url, path)
        extract_tar(path, self._raw_dir)
        os.unlink(path)

        raw_dir_path = osp.join(self._raw_dir, "nell_data")
        for root, dirs, files in os.walk(raw_dir_path, topdown=False):
            for file in files:
                if self._name in file:
                    shutil.move(osp.join(root, file), self._raw_dir)
        shutil.rmtree(raw_dir_path)

    def _process(self):
        objects = []
        for raw_file in self.raw_file_paths[:-1]:
            objects.append(pkl_read_file(raw_file))

        for i in range(len(objects) - 1):
            tmp = objects[i]
            tmp = tmp.todense() if hasattr(tmp, 'todense') else tmp
            objects[i] = torch.Tensor(tmp)

        x, tx, allx, y, ty, ally, graph = tuple(objects)
        test_index = read_txt_array(osp.join(self._raw_dir, f"ind.{self._name}.test.index"), dtype=torch.long)

        sorted_test_index = test_index.sort()[0]

        tx_ext = torch.zeros(len(graph) - allx.shape[0], x.shape[1])
        tx_ext[sorted_test_index - allx.shape[0]] = tx

        ty_ext = torch.zeros(len(graph) - ally.shape[0], y.shape[1])
        ty_ext[sorted_test_index - ally.shape[0]] = ty

        tx, ty = tx_ext, ty_ext

        # PyG creates feature vectors for relations here. Whereas the augmented features consume
        # too much memory to be stored in a dense matrix (numpy matrix), thus we ignore that.
        x = torch.cat([allx, tx], dim=0)
        x[test_index] = x[sorted_test_index]
        x = x.numpy()

        y = torch.cat([ally, ty], dim=0).max(dim=1)[1]
        y[test_index] = y[sorted_test_index]
        y = torch.LongTensor(y)

        adj = sp.coo_matrix(nx.adjacency_matrix(nx.from_dict_of_lists(graph)))
        row, col, edge_weight = adj.row, adj.col, adj.data

        num_node = y.shape[0]
        node_type = "knowledge"
        edge_type = "knowledge__to__knowledge"

        g = Graph(row, col, edge_weight, num_node, node_type, edge_type, x=x, y=y)
        with open(self.processed_file_paths, 'wb') as rf:
            try:
                pkl.dump(g, rf)
            except IOError as e:
                print(e)
                exit(1)

    def __generate_split(self, split):
        if split == "official":
            y = pkl_read_file(osp.join(self._raw_dir, f"ind.{self._name}.y"))
            test_idx = read_txt_array(osp.join(self._raw_dir, f"ind.{self._name}.test.index"), dtype=torch.long)
            train_idx = torch.arange(y.shape[0], dtype=torch.long)
            val_idx = torch.arange(y.shape[0], y.shape[0] + 500, dtype=torch.long)

        elif split == "random":
            raise NotImplementedError
        else:
            raise ValueError("Please input valid split pattern!")

        return train_idx, val_idx, test_idx



from sgl.models.base_model import BaseSGAPModel
from sgl.models.simple_models import IdenticalMapping
from sgl.operators.graph_op import LaplacianGraphOp
from sgl.operators.message_op import OverSmoothDistanceWeightedOp


class NAFS(BaseSGAPModel):
    def __init__(self, prop_steps, feat_dim, output_dim):
        super(NAFS, self).__init__(prop_steps, feat_dim, output_dim)

        self._pre_graph_op = LaplacianGraphOp(prop_steps, r=0.5)
        self._pre_msg_op = OverSmoothDistanceWeightedOp()
        self._base_model = IdenticalMapping()



import torch

from sgl.dataset import OgbnMag
from sgl.models.hetero import NARS_SIGN
from sgl.tasks import HeteroNodeClassification
from sgl.utils import GpuWithMaxFreeMem

dataset = OgbnMag("mag", "./")
predict_class = 'paper'
random_subgraph_num = 2
subgraph_edge_type_num = 1
model = NARS_SIGN(prop_steps=3, feat_dim=dataset.data.num_features[predict_class],
                  output_dim=dataset.data.num_classes[predict_class], hidden_dim=256, num_layers=2,
                  random_subgraph_num=random_subgraph_num, subgraph_edge_type_num=subgraph_edge_type_num)

device = torch.device(f"cuda:{0}" if torch.cuda.is_available() else "cpu")
test_acc = HeteroNodeClassification(dataset, predict_class, model, lr=0.001, weight_decay=0, epochs=200, device=device,
                                    train_batch_size=10000, eval_batch_size=10000).test_acc
